# -*- coding: utf-8 -*-

import json
import os
import sys
import tempfile

from string import Template

try:
    from unittest2 import TestCase
except ImportError:
    from unittest import TestCase

import six

from jinja2.utils import escape
from routr import route
from webob.response import Response
from webtest import TestApp

from rororo import GET, create_app, exceptions, manage
from rororo.exceptions import ImproperlyConfigured, RouteReversalError
from rororo.utils import absdir, force_unicode


DEBUG = True
STATIC_DIR = TEMPLATE_DIR = os.path.join(tempfile.gettempdir(), 'rororo')
TEMPLATE = '<h1>{{ var }}</h1>'
TEMPLATE_NAME = 'template.html'
TEMPLATE_WITH_GLOBALS = """<h1>Hello, world!</h1>

<h2>Links</h2>
<ul>
    <li><a href="{{ reverse("index") }}">Index view</a></li>
    <li><a href="{{ reverse("lambda") }}">Lambda view</a></li>
    <li><a href="{{ reverse("json") }}">JSON view</a></li>
    <li><a href="{{ reverse("server_error") }}">Server Error view</a></li>
</ul>

<h2>Settings</h2>
<ul>
    <li>APP_DIR: {{ settings.APP_DIR }}</li>
    <li>DEBUG: {{ settings.DEBUG }}</li>
    <li>JINJA_OPTIONS: {{ settings.JINJA_OPTIONS }}</li>
    <li>RENDERERS: {{ settings.RENDERERS }}</li>
    <li>TEMPLATE_DIR: {{ settings.TEMPLATE_DIR }}</li>
</ul>
"""

RENDERERS = (
    ('custom_json', 'rororo.tests.custom_json_renderer'),
    ('custom_template', 'rororo.tests.custom_template_renderer'),
    ('wrong', Response),
)
ROUTES = (
    GET('/', 'rororo.tests.index_view', name='index'),
    GET('/json', 'rororo.tests.json_view', name='json', renderer='json'),
    GET('/json/custom', 'rororo.tests.json_view', renderer='custom_json'),
    GET('/lambda', lambda: 'Hello, world!', name='lambda', renderer='text'),
    GET('/server-error-1',
        'rororo.tests.server_error_view',
        name='server_error'),
    GET('/server-error-2', 'rororo.tests.index_view', renderer='error'),
    GET('/server-error-3', 'rororo.tests.index_view', renderer='wrong'),
    GET('/template',
        'rororo.tests.template_view',
        name='template',
        renderer=TEMPLATE_NAME),
    GET('/template/custom',
        'rororo.tests.template_view',
        renderer='custom_template'),
)


class TestRororo(TestCase):

    def setUp(self):
        if not os.path.isdir(STATIC_DIR):
            os.mkdir(STATIC_DIR)

        self.template = template = os.path.join(TEMPLATE_DIR, TEMPLATE_NAME)

        with open(template, 'w+') as handler:
            handler.write(TEMPLATE)

    def tearDown(self):
        if hasattr(self, 'old_argv'):
            sys.argv = self.old_argv

        if hasattr(self, 'old_stdout'):
            sys.stdout = self.old_stdout

        if hasattr(self, 'old_stderr'):
            sys.stderr = self.old_stderr

    def test_create_app(self):
        app = create_app(__name__)
        self.assertTrue(callable(app), repr(app))

    def test_create_app_default_settings(self):
        app = create_app(routes=ROUTES)
        self.assertEqual(app.settings.APP_DIR,
                         os.getcwdu() if not six.PY3 else os.getcwd())
        self.assertFalse(app.settings.DEBUG)
        self.assertEqual(app.settings.JINJA_GLOBALS, {})
        self.assertEqual(app.settings.JINJA_FILTERS, {})
        self.assertEqual(app.settings.JINJA_OPTIONS, {})
        self.assertEqual(app.settings.PACKAGES, ())
        self.assertEqual(app.settings.RENDERERS, ())
        self.assertEqual(app.settings.STATIC_DIR, 'static')
        self.assertEqual(app.settings.TEMPLATE_DIR, 'templates')

    def test_create_app_improperly_configured(self):
        self.assertRaises(ImproperlyConfigured, create_app)
        self.assertRaises(ImproperlyConfigured, create_app, exceptions)
        self.assertRaises(ImproperlyConfigured, create_app, routes={})

    def test_create_app_multiple_apps(self):
        routes = ('', GET('/', index_view))
        text_app = create_app(routes=routes)

        routes = ('', GET('/', json_view, renderer='json'))
        json_app = create_app(routes=routes)

        self.assertNotEqual(text_app.reverse, json_app.reverse)
        self.assertNotEqual(text_app.routes, json_app.routes)
        self.assertNotEqual(text_app.settings, json_app.settings)

    def test_create_app_reverse(self):
        app = create_app(__name__)

        self.assertEqual(app.reverse('index'), '/')
        self.assertEqual(app.reverse('index'), '/')
        self.assertEqual(app.reverse('json'), '/json')
        self.assertEqual(app.reverse('server_error'), '/server-error-1')
        self.assertEqual(app.reverse('template'), '/template')
        self.assertRaises(RouteReversalError, app.reverse, 'does_not_exist')

        self.assertEqual(app.routes.reverse('index'), '/')
        self.assertEqual(app.routes.reverse('json'), '/json')
        self.assertEqual(app.routes.reverse('server_error'), '/server-error-1')
        self.assertEqual(app.routes.reverse('template'), '/template')
        self.assertRaises(RouteReversalError,
                          app.routes.reverse,
                          'does_not_exist')

    def test_jinja_renderer(self):
        dirname = os.path.abspath(os.path.dirname(__file__))

        with open(self.template, 'w+') as handler:
            handler.write(TEMPLATE_WITH_GLOBALS)

        app = TestApp(create_app(__name__))
        response = app.get('/template', status=200)

        self.assertIn('<a href="/">', response.text)
        self.assertIn('<a href="/json">', response.text)
        self.assertIn('<a href="/lambda">', response.text)
        self.assertIn('<a href="/server-error-1">', response.text)

        self.assertIn('APP_DIR: {0}'.format(dirname), response.text)
        self.assertIn('DEBUG: True', response.text)
        self.assertIn('JINJA_OPTIONS: {}', response.text)
        self.assertIn(
            'RENDERERS: {0}'.format(escape(RENDERERS)), response.text
        )
        self.assertIn(
            'TEMPLATE_DIR: {0}'.format(escape(TEMPLATE_DIR)), response.text
        )

        new_app = TestApp(create_app(routes=ROUTES, template_dir=TEMPLATE_DIR))
        response = new_app.get('/template', status=200)
        self.assertIn('DEBUG: False', response.text)
        self.assertIn('RENDERERS: ()', response.text)

        response = app.get('/template', status=200)
        self.assertIn('DEBUG: True', response.text)
        self.assertNotIn('RENDERERS: ()', response.text)

    def test_manage(self):
        self.old_argv = sys.argv
        self.old_stdout, self.old_stderr = sys.stdout, sys.stderr

        stdout, stderr = six.StringIO(), six.StringIO()
        sys.argv = [sys.argv[0]]
        sys.stdout, sys.stderr = stdout, stderr

        # Run manager without arguments
        app = create_app(__name__)

        try:
            manage(app)
        except SystemExit as err:
            self.assertEqual(err.code, 2)

        # Run clean_pyc management command
        sys.argv = [sys.argv[0], 'clean_pyc']

        try:
            manage(app)
        except SystemExit as err:
            self.assertEqual(err.code, 0)

        # Run print_settings management command
        sys.argv = [sys.argv[0], 'print_settings']

        try:
            manage(app)
        except SystemExit as err:
            self.assertEqual(err.code, 0)

        # Run undefined management command
        sys.argv = [sys.argv[0], 'does_not_exist']

        try:
            manage(app)
        except SystemExit as err:
            self.assertEqual(err.code, 2)

        # Run custom management command
        sys.argv = [sys.argv[0], 'custom_command']

        try:
            manage(app)
        except SystemExit as err:
            self.assertEqual(err.code, 2)

        try:
            manage(app, custom_command)
        except SystemExit as err:
            self.assertEqual(err.code, 0)

    def test_routing_and_renderers(self):
        app = TestApp(create_app(__name__))

        response = app.get('/', status=200)
        self.assertEqual(
            response.headers['Content-Type'], 'text/html; charset=UTF-8'
        )
        self.assertEqual(response.text, 'Hello, world!')

        response = app.get('/json', status=200)
        self.assertEqual(
            response.headers['Content-Type'], 'application/json; charset=UTF-8'
        )
        self.assertEqual(response.json, {'var': 'Hello, world!'})

        response = app.get('/json/custom', status=200)
        self.assertEqual(
            response.headers['Content-Type'], 'application/json; charset=UTF-8'
        )
        self.assertEqual(response.json, {'var': 'Hello, world!'})

        response = app.get('/lambda', status=200)
        self.assertEqual(
            response.headers['Content-Type'], 'text/plain; charset=UTF-8'
        )
        self.assertEqual(response.text, 'Hello, world!')

        response = app.get('/template', status=200)
        self.assertEqual(
            response.headers['Content-Type'], 'text/html; charset=UTF-8'
        )
        self.assertEqual(response.text, '<h1>Hello, world!</h1>')

        response = app.get('/template/custom', status=200)
        self.assertEqual(
            response.headers['Content-Type'], 'text/html; charset=UTF-8'
        )
        self.assertEqual(response.text, '<h1>Hello, world!</h1>')

        app.get('/does_not_exist.exe', status=404)
        app.get('/server-error-1', status=500)
        app.get('/server-error-2', status=500)
        app.get('/server-error-3', status=500)

    def test_server_error(self):
        app = TestApp(create_app(__name__))

        response = app.get('/server-error-1', status=500)
        self.assertIn('Test Exception', response.text)

        response = app.get('/server-error-2', status=500)
        self.assertIn("Renderer 'error' does not exist.", response.text)

        response = app.get('/server-error-3', status=500)
        self.assertIn('Renderer function has wrong args spec.', response.text)

    def test_server_error_debug(self):
        app = TestApp(create_app(debug=False, routes=ROUTES))
        response = app.get('/server-error-1', status=500)
        self.assertNotIn('Traceback', response.text)

        app = TestApp(create_app(debug=True, routes=ROUTES))
        response = app.get('/server-error-1', status=500)
        self.assertIn('Traceback', response.text)

    def test_static(self):
        app = TestApp(create_app(__name__))
        response = app.get('/static/{0}'.format(TEMPLATE_NAME), status=200)
        self.assertEqual(response.text, TEMPLATE)

    def test_static_routes(self):
        app = create_app(routes=ROUTES)
        routes = route(*ROUTES)
        self.assertNotEqual(len(routes.routes), len(app.routes.routes))

        app = create_app(debug=True, routes=ROUTES)
        self.assertNotEqual(len(routes.routes), len(app.routes.routes))

    def test_utils_absdir(self):
        self.assertEqual(absdir('user', '/Users'), '/Users/user')
        self.assertEqual(absdir('/Users/user', '/Users'), '/Users/user')

    def test_utils_force_unicode(self):
        u = str if six.PY3 else lambda value: unicode(value, 'utf-8')

        self.assertEqual(force_unicode('hello'), u('hello'))
        self.assertEqual(force_unicode('привет'), u('привет'))

        self.assertEqual(force_unicode(u('hello')), u('hello'))
        self.assertEqual(force_unicode(u('привет')), u('привет'))


def custom_command(app):
    print('Hello, world!')


def custom_json_renderer(data):
    return Response(json.dumps(data, indent=4),
                    content_type='application/json')


def custom_template_renderer(settings, renderer, data):
    return Response(Template('<h1>${var}</h1>').substitute(data))


def index_view():
    return 'Hello, world!'


def json_view():
    return {'var': 'Hello, world!'}


def server_error_view():
    raise Exception('Test Exception')


def template_view():
    return {'var': 'Hello, world!'}



"""
=================
rororo.exceptions
=================

All exceptions from WebOb, routr and rororo in one place.

"""

from string import Template

from routr import exc as routr_exceptions
from webob import exc as webob_exceptions


class HTTPServerError(webob_exceptions.HTTPServerError):
    """
    Modify server errors to print tracebacks in debug mode.
    """
    body_template_obj = Template("""${explanation}<br /><br />
<pre>${detail}</pre>
${html_comment}
""")


class ImproperlyConfigured(Exception):
    """
    Class for improperly configured errors.
    """


class NoRendererFound(ValueError):
    """
    Prettify message for "No renderer found" exception.
    """
    def __init__(self, renderer):
        message = 'Renderer {0!r} does not exist.'.format(renderer)
        super(NoRendererFound, self).__init__(message)


class ValidationError(Exception):
    """
    Class for validation errors.

    This is very basic class, no specific outputs or other things which
    happened somewhere else. Just put an error message and then print it in
    template for example.
    """


def inject(module):
    """
    Inject all exceptions from module to global namespace.
    """
    for attr in dir(module):
        if attr.startswith('_') or attr in globals():
            continue

        value = getattr(module, attr)

        if isinstance(value, type) and issubclass(value, Exception):
            globals()[value.__name__] = value


inject(routr_exceptions)
inject(webob_exceptions)



"""
============
rororo.utils
============

Common utilities to use in ``rororo`` and other projects.

"""

import os

import six


def absdir(dirname, base):
    """
    Prepend ``base`` path to ``dirname`` path if it relative.
    """
    if os.path.isabs(dirname):
        return dirname
    return os.path.abspath(os.path.join(base, dirname))


def force_unicode(value, encoding='utf-8', errors='ignore'):
    """
    Cast value to unicode representation.
    """
    if isinstance(value, six.text_type):
        return value
    return value.decode(encoding, errors)



"""
=============
rororo.static
=============

Extended static view which has support of package static directories.

"""

import os

from routr import GET
from routr.static import _ForceResponse
from webob.static import FileApp

from .schema import defaults


def make_static_view(dirnames):
    """
    Make actual static view.
    """
    def static_view(request, path):
        """
        Try to find requested ``path`` in ``dirnames`` sequence, if possible
        return file content to browser, else raise 404 error.
        """
        sequence = (dirnames
                    if isinstance(dirnames, (list, tuple))
                    else (dirnames, ))

        for dirname in sequence:
            filename = os.path.join(dirname, path)

            if os.path.isfile(filename):
                return _ForceResponse(FileApp(filename)(request))

        # If no static file found use app static dir to point about error,
        # not package static dir
        filename = os.path.join(dirnames[0], path)
        return _ForceResponse(FileApp(filename)(request))

    static_view.static_view = True
    return static_view


def static(prefix, dirnames=None, **kwargs):
    """
    Route for serving static assets.

    In comparison with default ``routr.static:static`` function ``static`` has
    support of multiple static directories.
    """
    kwargs['static_view'] = True
    path = kwargs.pop('path', None)

    if path:
        return GET(prefix,
                   defaults(path=path),
                   make_static_view(dirnames),
                   **kwargs)

    return GET('{0}/{{path:path}}'.format(prefix.rstrip('/')),
               make_static_view(dirnames),
               **kwargs)



"""
==========
rororo.app
==========

Create WSGI application and configure all necessary details.

Create application
==================

The main idea of ``rororo`` is an ability to fast construct and use WSGI apps
without Python OOP layer. And as ``rororo`` inspired from ``routr`` library
the next idea is reuse its routing system and made them a core of framework.

Routing
-------

You know, it maybe sounds very praised, but for me only necessary thing for
WSGI app is proper routing layer for connect requested path info with necessary
view function.

And next big question, are we really everytime need to pass ``request``
instance to our view functions? Sometime this is very big limitation and
sometime this limitation walks our views away from all other buisness logic.

Process renderers
=================

Custom renderers
----------------

Available settings
==================

APP_DIR
-------

Application directory. Needed if static or template directory has relative not
absolute path, then rel path joined with app directory to result.

By default: directory where settings module located or current working
directory.

DEBUG
-----

Debug mode. Main core feature of debug mode is support static files by adding
static route to list of all application routes.

By default: ``False``

JINJA_FILTERS
-------------

Dictionary with filters to use in Jinja templates.

By default: ``{}``

JINJA_GLOBALS
-------------

Dictionary with globals which send to each Jinja template.

By default: ``{}``

JINJA_OPTIONS
-------------

Options to initialize Jinja2 environment.

By default: ``{}``

PACKAGES
--------

Sequence with all additional packages which would be installed to current app.

Package is just a Python package compatible with rororo, same as reusable app
is for Django project or Flask blueprint is for Flask app.

Main reason of adding package to application is adding additional views,
templates and static directories to application.

.. important:: Right now it is concept and not a final solution.

By default: ``()``

PEP8_OPTIONS
------------

Keyword arguments passed to ``pep8.StyleGuide`` class on initialization when
checking PEP8 for current app and all its packages.

By default: ``{'statistics': True}``

RENDERERS
---------

Sequence with all custom renderers in format ``(test, func)``.

By default: ``()``

ROUTES
------

**Required**. All available routes for your app in list or tuple format. These
routes would be wrapped with ``routr.route`` function.

STATIC_DIR
----------

Directory where static files placed. Necessary only for debug mode and
easyfying static files support. If value is relpath it would be joined with app
directory.

By default: ``'static'``

STATIC_URL
----------

URL for static files route. Used only in debug mode.

If you want to have static files handling by WebOb in any mode - just add::

    static(STATIC_URL, STATIC_DIR, name='static')

route to list of all ``ROUTES``.

By default: ``'/static'``

TEMPLATE_DIR
------------

Directory with Jinja templates. If value is relpath it would be joined with app
directory.

By default: ``'templates'``

USE_WDB
-------

Wrap WSGI application in `WDB web-debugger <https://github.com/Kozea/wdb>`_ or
not.

By default: ``False``

WDB_OPTIONS
-----------

Keyword arguments passed to ``wdb.Wdb`` class init when wrapping original WSGI
application.

By default: ``{'start_disabled': True}``

"""

import copy
import inspect
import operator
import os
import traceback

import six

from jinja2 import Environment, FileSystemLoader
from routr import Endpoint, route
from routr.utils import import_string, inject_args
from webob.request import Request
from webob.response import Response

from .exceptions import (
    ImproperlyConfigured, HTTPException, HTTPServerError, NoMatchFound,
    NoRendererFound
)
from .static import static
from .utils import absdir, force_unicode


DEFAULT_RENDERERS = (
    ('json', lambda data: Response(json=data)),
    ('text', lambda data: Response(data, content_type='text/plain')),
    (lambda renderer: renderer is None, lambda data: Response(data)),
    (
        lambda renderer: renderer.endswith(JINJA_HTML_TEMPLATES),
        lambda settings, renderer, data: jinja_renderer(
            settings, renderer, data
        )
    )
)
DEFAULT_SETTINGS = (
    ('APP_DIR', None),
    ('DEBUG', False),
    ('JINJA_FILTERS', {}),
    ('JINJA_GLOBALS', {}),
    ('JINJA_OPTIONS', {}),
    ('PACKAGES', ()),
    ('PEP8_OPTIONS', {'statistics': True}),
    ('RENDERERS', ()),
    ('STATIC_DIR', 'static'),
    ('STATIC_URL', '/static'),
    ('TEMPLATE_DIR', 'templates'),
    ('USE_PEP8', False),
    ('USE_WDB', False),
    ('WDB_OPTIONS', {'start_disabled': True}),
)
JINJA_HTML_TEMPLATES = ('.htm', '.html', '.xhtml', '.xml')


def create_app(mixed=None, **kwargs):
    """
    Load settings from Python module and return valid WSGI application.

    Settings should contain ``ROUTES`` var, else ``ImproperlyConfigured`` error
    would be raised.

    ``ROUTES`` should be a ``tuple`` which would be wrapped with
    ``routr.route`` function.
    """
    def application(environ, start_response):
        """
        WSGI application. Place where all routr handling done.
        """
        # Initialize request, based on current environ
        request = Request(environ)

        # Read routes from current config and initialize them
        routes = get_routes(settings)

        # Start routing
        try:
            # Trace all available routes to find view to execute
            trace = routes(request)
            view = trace.target

            # If target is string - try to import function from its notation
            view = (import_string(view)
                    if isinstance(view, six.string_types)
                    else view)

            # Inject request into view args if necessary
            args, kwargs = trace.args, trace.kwargs
            args = inject_args(view, args, request=request)

            # Run view function
            response = view(*args, **kwargs)

            # If response is not WebOb's instance - process it with all
            # available renderers
            if not isinstance(response, Response):
                renderer = trace.annotation('renderer')
                response = process_renderer(settings, renderer, response)
        # Process exceptions
        except Exception as err:
            # If WDB web debugger used - raise original exception
            if settings.USE_WDB:
                raise

            # No route found in list of available ones
            if isinstance(err, NoMatchFound):
                response = err.response
            # HTTP exception happened
            elif isinstance(err, HTTPException):
                response = err
            # Wrap any other exception to "Server Error"
            else:
                message = traceback.format_exc() if settings.DEBUG else None
                response = HTTPServerError(message)

        # And finally return response to WSGI server
        return response(environ, start_response)

    # Load default setttings from predefined values
    settings = type('Settings', (object, ), dict(DEFAULT_SETTINGS))()

    # Make able to load settings from pathes, not only from previously imported
    # modules
    if mixed:
        all_settings = (import_string(mixed)
                        if isinstance(mixed, six.string_types)
                        else mixed)
    # But also have ability to load settings from keyword arguments to easify
    # creating applications for testing
    else:
        all_settings = None
        kwargs = dict(zip(
            map(operator.methodcaller('upper'), kwargs.keys()),
            kwargs.values()
        ))
        settings.__dict__.update(kwargs)

    # Only uppercased keys would be stored
    for attr in dir(all_settings):
        if attr.startswith('_') or not attr.isupper():
            continue
        setattr(settings, attr, getattr(all_settings, attr))

    # If ``APP_DIR`` not specified - set it to dir, where settings module
    # placed or current work directory
    if not settings.APP_DIR:
        if hasattr(all_settings, '__file__'):
            dirname = os.path.abspath(os.path.dirname(all_settings.__file__))
        else:
            dirname = os.getcwdu() if not six.PY3 else os.getcwd()

        settings.APP_DIR = dirname

    # Path should be an unicode
    settings.APP_DIR = force_unicode(settings.APP_DIR)

    # Setup initial template and static directories sequences
    settings._PACKAGE_DIRS = []
    settings._STATIC_DIRS = [absdir(settings.STATIC_DIR, settings.APP_DIR)]
    settings._TEMPLATE_DIRS = [absdir(settings.TEMPLATE_DIR, settings.APP_DIR)]

    # Now it's time to load all packages if any and modify settings a bit if
    # necessary (settings.PACKAGES is not an empty sequence)
    packages = register_packages(settings)

    # And finally do some validation, check that routes properly configured and
    # if yes save them in application as well as settings
    routes = get_routes(settings)

    setattr(application, 'packages', packages)
    setattr(application, 'reverse', routes.reverse)
    setattr(application, 'routes', routes)
    setattr(application, 'settings', settings)

    # Return valid WSGI application
    return application


def get_routes(settings):
    """
    Read routes from configuration and wrap it with ``routr.route`` function.
    """
    # Settings should contain ``ROUTES``
    try:
        routes = settings.ROUTES
    except AttributeError:
        raise ImproperlyConfigured('Please supply ROUTES setting. This '
                                   'setting is required.')

    # Routes should be a list or a tuple
    if not isinstance(routes, (list, tuple)):
        raise ImproperlyConfigured('ROUTES should be a list or tuple.')

    # Add static view to routes
    routes = list(routes)
    routes.insert(
        1 if isinstance(routes[0], six.string_types) else 0,
        static(settings.STATIC_URL, settings._STATIC_DIRS, name='static')
    )

    # And finally initialize routes with wrapping to ``route`` function
    return route(*routes)


def jinja_autoescape(filename):
    """
    Should we acitvate autoescaping for given filename or not.
    """
    if not filename:
        return False
    return filename.endswith(JINJA_HTML_TEMPLATES)


def jinja_env(settings):
    """
    Prepare Jinja environment.
    """
    # Prepare Jinja options
    options = copy.deepcopy(settings.JINJA_OPTIONS)
    options.setdefault('autoescape', jinja_autoescape)
    options.setdefault('loader', FileSystemLoader(settings._TEMPLATE_DIRS))

    # And make Jinja environment
    env = Environment(**options)

    # Populate all filters and globals from settings
    for key, value in six.iteritems(settings.JINJA_FILTERS):
        env.filters[key] = value

    for key, value in six.iteritems(settings.JINJA_GLOBALS):
        env.globals[key] = value

    # Remove ability of overwriting reverse and settings globals
    env.globals['reverse'] = get_routes(settings).reverse
    env.globals['settings'] = settings

    return env


def jinja_renderer(settings, filename, data):
    """
    Render template from ``filename`` using Jinja template engine.
    """
    if not isinstance(data, dict):
        raise ValueError('Unable to render template without proper context. '
                         'Passed context: {0!r}'.format(data))

    template = jinja_env(settings).get_template(filename)
    return Response(template.render(**data))


def match_renderer(key, renderer):
    """
    Returns True if ``key`` matches ``renderer``. ``key`` could be callable or
    static value.
    """
    return (callable(key) and key(renderer)) or \
           (not callable(key) and key == renderer)


def process_renderer(settings, renderer, data):
    """
    Extract ``renderer`` metadata from trace and run all available renderers to
    convert data to valid response.
    """
    renderers = tuple(DEFAULT_RENDERERS) + tuple(settings.RENDERERS)

    for key, mixed in renderers:
        if match_renderer(key, renderer):
            is_class = False

            if isinstance(mixed, six.string_types):
                func = import_string(mixed)
            elif inspect.isclass(mixed):
                func = mixed.__init__
                is_class = True
            else:
                func = mixed

            args_count = len(inspect.getargspec(func).args or ())
            args_count = args_count - 1 if is_class else args_count
            func = mixed if is_class else func

            if args_count == 1:
                return func(data)
            elif args_count == 3:
                return func(settings, renderer, data)
            else:
                raise ValueError('Renderer function has wrong args spec.')

    raise NoRendererFound(renderer)


def register_packages(settings):
    """
    If ``settings.PACKAGES`` is not an empty list/tuple - extend routes,
    template and static directories and jinja globals/filters from package
    settings if any.
    """
    def append_settings(settings, package_settings, package_name, package_dir):
        """
        Append routes, template and static directories to global settings.
        """
        if not isinstance(settings.ROUTES, list):
            settings.ROUTES = list(settings.ROUTES)

        static_dir = getattr(package_settings,
                             'STATIC_DIR',
                             default_settings['STATIC_DIR'])

        template_dir = getattr(package_settings,
                               'TEMPLATE_DIR',
                               default_settings['TEMPLATE_DIR'])

        settings._STATIC_DIRS.append(absdir(static_dir, package_dir))
        settings._TEMPLATE_DIRS.append(absdir(template_dir, package_dir))

        if hasattr(package_settings, 'JINJA_GLOBALS'):
            settings.JINJA_GLOBALS.update(package_settings.JINJA_GLOBALS)

        if hasattr(package_settings, 'JINJA_FILTERS'):
            settings.JINJA_FILTERS.update(package_settings.JINJA_FILTERS)

        if hasattr(package_settings, 'ROUTES'):
            settings.ROUTES.extend(
                safe_include(package_settings.ROUTES, package_name)
            )

    def safe_include(routes, package_name):
        """
        Safe include of routes from packages by wrapping them to route.
        """
        routes = list(routes)
        base_pattern = routes.pop(0)

        if isinstance(base_pattern, six.string_types):
            base_pattern = base_pattern.rstrip('/')
        else:
            routes.insert(0, base_pattern)
            base_pattern = None

        for endpoint in routes:
            if isinstance(endpoint, Endpoint) and endpoint.name:
                endpoint.name = ':'.join((package_name, endpoint.name))

                if base_pattern:
                    endpoint.pattern.pattern = (
                        ''.join((base_pattern, endpoint.pattern.pattern))
                    )

        return tuple(routes)

    # Initial vars
    default_settings = dict(DEFAULT_SETTINGS)
    packages = []

    # For each package, try to load it and load its settings
    for package_name in settings.PACKAGES:
        package = import_string(package_name)
        package_dir = force_unicode(
            os.path.abspath(os.path.dirname(package.__file__))
        )

        # Package could be without settings too. This just means that we don't
        # need to do any modifications with current settings
        package_settings_module = '{0}.settings'.format(package_name)

        try:
            package_settings = import_string(package_settings_module)
        except ImportError:
            pass
        else:
            args = (settings, package_settings, package_name, package_dir)
            append_settings(*args)

        settings._PACKAGE_DIRS.append(package_dir)
        packages.append(package_name)

    return tuple(packages)



"""
==============
rororo.manager
==============

Support of custom management commands for rororo framework.

"""

import copy
import inspect
import operator
import subprocess
import sys
import types

from argparse import ArgumentParser
from wsgiref.simple_server import make_server

import server_reloader
import six

from routr.utils import import_string


DEFAULT_HOST = '0.0.0.0'
DEFAULT_PORT = 8000


def manage(app, *commands):
    """
    Find all possible management commands from functions in current module and
    add them as sub-commands to main argument parser.

    You also could to add more custom management commands as positional
    arguments while calling this function.

    .. note:: Each management function should get ``app`` as first positional
       arg. All other arguments should be optional.
    """
    # Initialize parser
    parser = ArgumentParser(description='rororo management commands.')
    subparsers = parser.add_subparsers(
        description='List of available commands.',
        title='commands'
    )

    # Find all available management commands and add they as parser sub-command
    data = list(copy.copy(globals()).items())
    ignore = ('manage', )

    for func in commands:
        data.append((func.func_name if not six.PY3 else func.__name__, func))

    data = sorted(data, key=operator.itemgetter(0))

    for key, value in data:
        # Only functions from current module would be available as sub-commands
        if (
            key.startswith('_') or key in ignore or
            not isinstance(value, types.FunctionType) or
            (value.__module__ != __name__ and not value in commands)
        ):
            continue

        # Add sub-command to main parser
        kwargs = {}
        if value.__doc__:
            kwargs['help'] = value.__doc__.replace('\n', ' ').strip()
        subparser = subparsers.add_parser(key, **kwargs)

        # Read function args spec to add necessary arguments to subcommand
        spec = inspect.getargspec(value)

        args, defaults = spec.args or (), spec.defaults or ()
        defaults_start = len(args) - len(defaults)

        for i, arg in enumerate(args):
            if arg == 'app' or arg.startswith('_'):
                continue

            kwargs = {}

            if i >= defaults_start:
                default = kwargs['default'] = defaults[i - defaults_start]

                if isinstance(default, bool):
                    kwargs['action'] = \
                        'store_true' if not default else 'store_false'

                    if default:
                        arg = 'no-{0}'.format(arg)

                kwargs['help'] = 'By default: {0}'.format(default)

            arg = arg.replace('_', '-')
            subparser.add_argument('--{0}'.format(arg), **kwargs)

        # Run function
        subparser.set_defaults(func=value)

    # Parse arguments from command line
    no_args = len(sys.argv) == 1

    if not six.PY3 and no_args:
        old_stdout, old_stderr = sys.stdout, sys.stderr
        sys.stdout, sys.stderr = six.StringIO(), six.StringIO()

    try:
        args = parser.parse_args(sys.argv[1:])
    except SystemExit as err:
        if err.code:
            if not six.PY3 and no_args:
                sys.stdout, sys.stderr = old_stdout, old_stderr
                parser.print_help()
            sys.exit(err.code)
        raise err
    else:
        if six.PY3 and no_args:
            parser.print_help()
            sys.exit(2)

    # Run necessary function if everything ok
    kwargs = dict([
        (key.replace('-', '_').replace('no_', ''), value)
        for key, value in args._get_kwargs()
        if key != 'func'
    ])
    return args.func(app, **kwargs)


def clean_pyc(app):
    """
    Remove all *.pyc files which contains in APP_DIR.
    """
    cmd = 'find "{0}" -name "*.pyc" -exec rm -rf {{}} \;'.\
          format(app.settings.APP_DIR)
    subprocess.call(cmd, shell=True)


def pep8(app, _return_report=True):
    """
    Run PEP8 for all Python files in app directory and all available package
    directories.
    """
    # Nothing to do if PEP8 support disabled in settings
    if not app.settings.USE_PEP8:
        print('PEP8 checks disabled for current app. To enable define '
              '"USE_PEP8 = True" if app settings.')
        sys.exit(0)

    # Import PEP8 style guide
    guide_klass = import_string('pep8.StyleGuide')

    # Configure PEP8 style guide
    options = copy.deepcopy(app.settings.PEP8_OPTIONS)
    options['paths'] = [app.settings.APP_DIR] + app.settings._PACKAGE_DIRS

    # Initialize style guide and run files check
    guide = guide_klass(**options)
    report = guide.check_files()

    # If necessary - print statistics
    if _return_report:
        return report

    if guide.options.statistics:
        report.print_statistics()

    # If errors happened - show its number or just exit from function
    if report.total_errors:
        if guide.options.count:
            print >> sys.stderr, report.total_errors
        sys.exit(1)


def print_settings(app):
    """
    Print application settings as key => value lines.
    """
    for attr in dir(app.settings):
        if attr.startswith('_'):
            continue
        print('{0} = {1!r}'.format(attr, getattr(app.settings, attr)))


def runserver(app, host=DEFAULT_HOST, port=DEFAULT_PORT, autoreload=True):
    """
    Run Python's standard simple WSGI server with ability to autoreload on file
    changes.
    """
    def before_reload():
        """
        Show message before server reloaded.
        """
        print('Reloading server...\n')

    def run_server():
        """
        Run simple WSGI server.

        Server should be with or without WDB web debugger and run PEP8
        checkings before each run.
        """
        if app.settings.USE_PEP8:
            report = pep8(app, True)

            if report.total_errors:
                print >> sys.stderr, (
                    '\nPEP8 check resulted {0} error(s). Please, fix errors '
                    'before run development server or disable PEP8 check ups '
                    'in app settings.'.format(report.total_errors)
                )
                sys.exit(1)

        wdb_app = None

        if app.settings.USE_WDB:
            wdb_klass = import_string('wdb:Wdb')
            wdb_app = wdb_klass(app, **app.settings.WDB_OPTIONS)

        server = make_server(host, int(port), wdb_app or app)

        try:
            print('Starting server at http://{0}:{1}/'.format(host, port))
            server.serve_forever()
        except KeyboardInterrupt:
            print

    if autoreload:
        server_reloader.main(run_server, before_reload=before_reload)
    else:
        run_server()


def test(app, tests=None):
    """
    Run all or only specified tests from APP_DIR.
    """



"""
======
rororo
======

Functional nano web-framework built on top of `WebOb <http://webob.org/>`_,
`routr <http://routr.readthedocs.org>`_ and `Jinja <http://jinja.pocoo.org>`_.

"""

from routr import (
    DELETE, GET, HEAD, OPTIONS, PATCH, POST, PUT, TRACE, include, plug, route
)
from routrschema import RequestParams, form, json_body, qs
from schemify import ValidationError, anything, opt, validate
from webob.request import Request
from webob.response import Response

from .app import create_app
from .manager import manage
from .static import static


__author__ = 'Igor Davydenko'
__framework__ = 'rororo'
__license__ = 'BSD License'
__version__ = '0.1-alpha'



"""
=============
rororo.schema
=============

Additional schemas (guards) to routr's: form, qs and json_body.

"""

import copy


class defaults(object):
    """
    Simple schema to pass default values to view.
    """
    def __init__(self, **kwargs):
        """
        Initialize object with any keyword arguments.
        """
        self.__dict__.update(kwargs)

    def __call__(self, request, trace):
        """
        Update trace kwargs with defaults values.
        """
        trace.kwargs.update(copy.deepcopy(self.__dict__))



#!/usr/bin/env python

import os
import re
import sys

from setuptools import setup

DIRNAME = os.path.abspath(os.path.dirname(__file__))
rel = lambda *parts: os.path.abspath(os.path.join(DIRNAME, *parts))

README = open(rel('README.rst')).read()
INIT_PY = open(rel('rororo', '__init__.py')).read()
VERSION = re.findall("__version__ = '([^']+)'", INIT_PY)[0]


setup(
    name='rororo',
    version=VERSION,
    description=('Functional nano web-framework built on top of WebOb, routr '
                 'and Jinja'),
    long_description=README,
    author='Igor Davydenko',
    author_email='playpauseandstop@gmail.com',
    url='http://github.com/playpauseandstop/rororo',
    packages=[
        'rororo',
    ],
    install_requires=list(filter(None, [
        'Jinja2==2.6',
        'WebOb>=1.2.3',
        'argparse==1.2.1' if sys.version_info[:2] < (2, 7) else None,
        'routr>=0.7.1',
        'routr-schema>=0.1',
        'schemify>=0.1',
        'server-reloader>=0.1.2',
        'unittest2==0.5.1' if sys.version_info[:2] < (2, 7) else None,
    ])),
    platforms='any',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Utilities',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'License :: OSI Approved :: BSD License',
    ],
    keywords='nano web framework webob routr jinja',
    license='BSD License',
)
